@@ -1,10 +1,10 @@
 /******************************************************************************
- * $Id: tif_ovrcache.c,v 1.4 2005/06/07 14:26:04 dron Exp $
+ * $Id: tif_ovrcache.c,v 1.7 2006/03/25 17:52:37 joris Exp $
  *
  * Project:  TIFF Overview Builder
  * Purpose:  Library functions to maintain two rows of tiles or two strips
  *           of data for output overviews as an output cache. 
- * Author:   Frank Warmerdam, warmerda@home.com
+ * Author:   Frank Warmerdam, warmerdam@pobox.com
  *
  ******************************************************************************
  * Copyright (c) 2000, Frank Warmerdam
@@ -107,8 +107,9 @@ TIFFOvrCache *TIFFCreateOvrCache( TIFF *hTIFF, int nDirOffset )
     if( psCache->pabyRow1Blocks == NULL
         || psCache->pabyRow2Blocks == NULL )
     {
-        TIFFError( hTIFF->tif_name,
-                   "Can't allocate memory for overview cache." );
+		TIFFErrorExt( hTIFF->tif_clientdata, hTIFF->tif_name,
+					  "Can't allocate memory for overview cache." );
+        /* TODO: use of TIFFError is inconsistent with use of fprintf in addtiffo.c, sort out */
         return NULL;
     }
 
@@ -136,11 +137,12 @@ static void TIFFWriteOvrRow( TIFFOvrCache * psCache )
     int		nRet, iTileX, iTileY = psCache->nBlockOffset;
     unsigned char *pabyData;
     uint32	nBaseDirOffset;
-    
+    uint32 RowsInStrip;
+
 /* -------------------------------------------------------------------- */
-/*	If the output cache is multi-byte per sample, and the file	*/
-/*	being written to is of a different byte order than the current	*/
-/*	platform, we will need to byte swap the data. 			*/
+/*      If the output cache is multi-byte per sample, and the file      */
+/*      being written to is of a different byte order than the current  */
+/*      platform, we will need to byte swap the data.                   */
 /* -------------------------------------------------------------------- */
     if( TIFFIsByteSwapped(psCache->hTIFF) )
     {
@@ -168,71 +170,72 @@ static void TIFFWriteOvrRow( TIFFOvrCache * psCache )
 /* -------------------------------------------------------------------- */
 /*      Write blocks to TIFF file.                                      */
 /* -------------------------------------------------------------------- */
-    for( iTileX = 0; iTileX < psCache->nBlocksPerRow; iTileX++ )
-    {
-        int nTileID;
+	for( iTileX = 0; iTileX < psCache->nBlocksPerRow; iTileX++ )
+	{
+		int nTileID;
 
-        if (psCache->nPlanarConfig == PLANARCONFIG_SEPARATE)
-        {
-            int iSample;
+		if (psCache->nPlanarConfig == PLANARCONFIG_SEPARATE)
+		{
+			int iSample;
 
-            for( iSample = 0; iSample < psCache->nSamples; iSample++ )
-            {
-                pabyData = TIFFGetOvrBlock( psCache, iTileX, iTileY, iSample );
+			for( iSample = 0; iSample < psCache->nSamples; iSample++ )
+			{
+				pabyData = TIFFGetOvrBlock( psCache, iTileX, iTileY, iSample );
 
-                if( psCache->bTiled )
-                {
-                    nTileID =
-                        TIFFComputeTile( psCache->hTIFF,
-                                         iTileX * psCache->nBlockXSize,
-                                         iTileY * psCache->nBlockYSize,
-                                         0, (tsample_t) iSample );
-                    TIFFWriteEncodedTile( psCache->hTIFF, nTileID, 
-                                          pabyData,
-                                          TIFFTileSize(psCache->hTIFF) );
-                }
-                else
-                {
-                    nTileID =
-                        TIFFComputeStrip( psCache->hTIFF,
-                                          iTileY * psCache->nBlockYSize,
-                                          (tsample_t) iSample );
+				if( psCache->bTiled )
+				{
+					nTileID = TIFFComputeTile( psCache->hTIFF,
+					    iTileX * psCache->nBlockXSize,
+					    iTileY * psCache->nBlockYSize,
+					    0, (tsample_t) iSample );
+					TIFFWriteEncodedTile( psCache->hTIFF, nTileID,
+					    pabyData,
+					    TIFFTileSize(psCache->hTIFF) );
+				}
+				else
+				{
+					nTileID = TIFFComputeStrip( psCache->hTIFF,
+					    iTileY * psCache->nBlockYSize,
+					    (tsample_t) iSample );
+					RowsInStrip=psCache->nBlockYSize;
+					if ((iTileY+1)*psCache->nBlockYSize>psCache->nYSize)
+						RowsInStrip=psCache->nYSize-iTileY*psCache->nBlockYSize;
+					TIFFWriteEncodedStrip( psCache->hTIFF, nTileID,
+					    pabyData,
+					    TIFFVStripSize(psCache->hTIFF,RowsInStrip) );
+				}
+			}
 
-                    TIFFWriteEncodedStrip( psCache->hTIFF, nTileID,
-                                           pabyData,
-                                           TIFFStripSize(psCache->hTIFF) );
-                }
-            }
+		}
+		else
+		{
+			pabyData = TIFFGetOvrBlock( psCache, iTileX, iTileY, 0 );
 
-        }
-        else
-        {
-            pabyData = TIFFGetOvrBlock( psCache, iTileX, iTileY, 0 );
-
-            if( psCache->bTiled )
-            {
-                nTileID =
-                    TIFFComputeTile( psCache->hTIFF,
-                                     iTileX * psCache->nBlockXSize,
-                                     iTileY * psCache->nBlockYSize,
-                                     0, 0 );
-                TIFFWriteEncodedTile( psCache->hTIFF, nTileID, 
-                                      pabyData,
-                                      TIFFTileSize(psCache->hTIFF) );
-            }
-            else
-            {
-                nTileID =
-                    TIFFComputeStrip( psCache->hTIFF,
-                                      iTileY * psCache->nBlockYSize,
-                                      0 );
-
-                TIFFWriteEncodedStrip( psCache->hTIFF, nTileID,
-                                       pabyData,
-                                       TIFFStripSize(psCache->hTIFF) );
-            }
-        }
-    }
+			if( psCache->bTiled )
+			{
+				nTileID = TIFFComputeTile( psCache->hTIFF,
+				    iTileX * psCache->nBlockXSize,
+				    iTileY * psCache->nBlockYSize,
+				    0, 0 );
+				TIFFWriteEncodedTile( psCache->hTIFF, nTileID,
+				    pabyData,
+				    TIFFTileSize(psCache->hTIFF) );
+			}
+			else
+			{
+				nTileID = TIFFComputeStrip( psCache->hTIFF,
+				    iTileY * psCache->nBlockYSize,
+				    0 );
+				RowsInStrip=psCache->nBlockYSize;
+				if ((iTileY+1)*psCache->nBlockYSize>psCache->nYSize)
+					RowsInStrip=psCache->nYSize-iTileY*psCache->nBlockYSize;
+				TIFFWriteEncodedStrip( psCache->hTIFF, nTileID,
+				    pabyData,
+				    TIFFVStripSize(psCache->hTIFF,RowsInStrip) );
+			}
+		}
+	}
+	/* TODO: add checks on error status return of TIFFWriteEncodedTile and TIFFWriteEncodedStrip */
 
 /* -------------------------------------------------------------------- */
 /*      Rotate buffers.                                                 */
@@ -249,20 +252,23 @@ static void TIFFWriteOvrRow( TIFFOvrCache * psCache )
 /*      Restore access to original directory.                           */
 /* -------------------------------------------------------------------- */
     TIFFFlush( psCache->hTIFF );
-    
+    /* TODO: add checks on error status return of TIFFFlush */
     TIFFSetSubDirectory( psCache->hTIFF, nBaseDirOffset );
+    /* TODO: add checks on error status return of TIFFSetSubDirectory */
 }
 
 /************************************************************************/
 /*                          TIFFGetOvrBlock()                           */
 /************************************************************************/
 
+/* TODO: make TIFF_Downsample handle iSample offset, so that we can
+ * do with a single TIFFGetOvrBlock and no longer need TIFFGetOvrBlock_Subsampled */
 unsigned char *TIFFGetOvrBlock( TIFFOvrCache *psCache, int iTileX, int iTileY,
                                 int iSample )
 
 {
     int		nRowOffset;
-    
+
     if( iTileY > psCache->nBlockOffset + 1 )
         TIFFWriteOvrRow( psCache );
 
@@ -285,6 +291,33 @@ unsigned char *TIFFGetOvrBlock( TIFFOvrCache *psCache, int iTileX, int iTileY,
         return psCache->pabyRow2Blocks + nRowOffset;
 }
 
+/************************************************************************/
+/*                     TIFFGetOvrBlock_Subsampled()                     */
+/************************************************************************/
+
+unsigned char *TIFFGetOvrBlock_Subsampled( TIFFOvrCache *psCache, 
+                                           int iTileX, int iTileY )
+
+{
+    int		nRowOffset;
+
+    if( iTileY > psCache->nBlockOffset + 1 )
+        TIFFWriteOvrRow( psCache );
+
+    assert( iTileX >= 0 && iTileX < psCache->nBlocksPerRow );
+    assert( iTileY >= 0 && iTileY < psCache->nBlocksPerColumn );
+    assert( iTileY >= psCache->nBlockOffset
+            && iTileY < psCache->nBlockOffset+2 );
+    assert( psCache->nPlanarConfig != PLANARCONFIG_SEPARATE );
+
+    nRowOffset = iTileX * psCache->nBytesPerBlock;
+
+    if( iTileY == psCache->nBlockOffset )
+        return psCache->pabyRow1Blocks + nRowOffset;
+    else
+        return psCache->pabyRow2Blocks + nRowOffset;
+}
+
 /************************************************************************/
 /*                        TIFFDestroyOvrCache()                         */
 /************************************************************************/
